@@ -1,5 +1,9 @@
 # -*- coding: utf-8 -*-
 """
+Extends the :class:`CategoryTree` class in the :mod:`kwcoco.category_tree`
+module with torch methods for computing hierarchical losses / decisions.
+
+
 Notes from YOLO-9000:
     * perform multiple softmax operations over co-hyponyms
     * we compute the softmax over all sysnsets that are hyponyms of the same concept
@@ -9,491 +13,25 @@ Notes from YOLO-9000:
     hyponymn - a word of more specific meaning than a general or
         superordinate term applicable to it. For example, spoon is a
         hyponym of cutlery.
-
-Ignore:
-    class_energy.sum(dim=1)
-    cond_probs.sum(dim=1)
-    class_probs.sum(dim=1)
-
-    import kwplot
-    kwplot.autompl()
-    import graphid
-    graphid.util.show_nx(self.graph)
-
-    dset = coco_dataset.CocoDataset.demo()
 """
 from __future__ import absolute_import, division, print_function, unicode_literals
 import torch
 import kwarray
 import functools
-import itertools as it
 import networkx as nx
 import ubelt as ub
 import torch.nn.functional as F
 import numpy as np
+from kwcoco import CategoryTree as KWCOCO_CategoryTree  # raw category tree
+
+__all__ = ['CategoryTree']
 
 
-class CategoryTree(ub.NiceRepr):
+class Mixin_CategoryTree_Torch:
     """
-    Helps compute softmaxes and probabilities for tree-based categories
-    where a directed edge (A, B) represents that A is a superclass of B.
-
-    TODO:
-        - [ ] separate the category tree from the method used to
-              compute the hierarchical softmax
-
-        - [ ] I think its ok to keep the decision function here.
-
-        - [ ] Should we not use an implicit root by default?
-
-    Example:
-        >>> from ndsampler.category_tree import *
-        >>> graph = nx.from_dict_of_lists({
-        >>>     'background': [],
-        >>>     'foreground': ['animal'],
-        >>>     'animal': ['mammal', 'fish', 'insect', 'reptile'],
-        >>>     'mammal': ['dog', 'cat', 'human', 'zebra'],
-        >>>     'zebra': ['grevys', 'plains'],
-        >>>     'grevys': ['fred'],
-        >>>     'dog': ['boxer', 'beagle', 'golden'],
-        >>>     'cat': ['maine coon', 'persian', 'sphynx'],
-        >>>     'reptile': ['bearded dragon', 't-rex'],
-        >>> }, nx.DiGraph)
-        >>> self = CategoryTree(graph)
-        >>> class_energy = torch.randn(3, len(self))
-        >>> class_probs = self.hierarchical_softmax(class_energy, dim=1)
-        >>> print(self)
-        <CategoryTree(nNodes=22, maxDepth=6, maxBreadth=4...)>
+    Mixin methods for CategoryTree that specifically relate to computing
+    normalized probabilities.
     """
-    def __init__(self, graph=None):
-        """
-        Args:
-            graph (nx.DiGraph):
-                either the graph representing a category hierarchy
-        """
-        if graph is None:
-            graph = nx.DiGraph()
-        else:
-            if len(graph) > 0:
-                if not nx.is_directed_acyclic_graph(graph):
-                    raise ValueError('The category graph must a DAG')
-                if not nx.is_forest(graph):
-                    raise ValueError('The category graph must be a forest')
-            if not isinstance(graph, nx.Graph):
-                raise TypeError('Input to CategoryTree must be a networkx graph not {}'.format(type(graph)))
-        self.graph = graph  # :type: nx.Graph
-        # Note: nodes are class names
-        self.id_to_node = None
-        self.node_to_id = None
-        self.node_to_idx = None
-        self.idx_to_node = None
-        self.idx_groups = None
-
-        self._build_index()
-
-    def copy(self):
-        new = self.__class__(self.graph.copy())
-        return new
-
-    @classmethod
-    def from_mutex(cls, nodes):
-        """
-        Args:
-            nodes (List[str]): or a list of class names (in which case they
-                will all be assumed to be mutually exclusive)
-
-        Example:
-            >>> print(CategoryTree.from_mutex(['a', 'b', 'c']))
-            <CategoryTree(nNodes=3, maxDepth=1, maxBreadth=3...)>
-        """
-        nodes = list(nodes)
-        graph = nx.DiGraph()
-        graph.add_nodes_from(nodes)
-        start = 0
-        if 'background' in graph.nodes:
-            # hack
-            graph.nodes['background']['id'] = 0
-            start = 1
-
-        for i, node in enumerate(nodes, start=start):
-            graph.nodes[node]['id'] = graph.nodes[node].get('id', i)
-
-        return cls(graph)
-
-    @classmethod
-    def from_json(cls, state):
-        self = cls()
-        self.__setstate__(state)
-        return self
-
-    @classmethod
-    def from_coco(cls, categories):
-        """
-        Create a CategoryTree object from coco categories
-
-        Args:
-            List[Dict]: list of coco-style categories
-        """
-        graph = nx.DiGraph()
-        for cat in categories:
-            graph.add_node(cat['name'], **cat)
-            if 'supercategory' in cat:
-                graph.add_edge(cat['supercategory'], cat['name'])
-        self = cls(graph)
-        return self
-
-    @classmethod
-    def coerce(cls, data):
-        """
-        Attempt to coerce data as a CategoryTree object.
-
-        This is primarily useful for when the software stack depends on
-        categories being represent
-
-        This will work if the input data is a specially formatted json dict, a
-        list of mutually exclusive classes, or if it is already a CategoryTree.
-        Otherwise an error will be thrown.
-
-        Args:
-            data (object): a known representation of a category tree.
-
-        Returns:
-            CategoryTree: self
-
-        Raises:
-            TypeError - if the input format is unknown
-        """
-        if isinstance(data, int):
-            # An integer specifies the number of classes.
-            self = cls.from_mutex(
-                ['class_{}'.format(i + 1) for i in range(data)])
-        elif isinstance(data, dict):
-            # A dictionary is assumed to be in a special json format
-            self = cls.from_json(data)
-        elif isinstance(data, list):
-            # A list is assumed to be a list of class names
-            self = cls.from_mutex(data)
-        elif isinstance(data, nx.DiGraph):
-            # A nx.DiGraph should represent the category tree
-            self = cls(data)
-        elif isinstance(data, cls):
-            # If data is already a CategoryTree, do nothing and just return it
-            self = data
-        else:
-            raise TypeError('Unknown type {}: {!r}'.format(type(data), data))
-        return self
-
-    @classmethod
-    def cast(cls, data):
-        import warnings
-        warnings.warn('cast is deprecated use coerce')
-        return cls.coerce(data)
-
-    @ub.memoize_property
-    def id_to_idx(self):
-        """
-        Example:
-            >>> import ndsampler
-            >>> self = ndsampler.CategoryTree.demo()
-            >>> self.id_to_idx[1]
-        """
-        return _calldict({cid: self.node_to_idx[node]
-                         for cid, node in self.id_to_node.items()})
-
-    @ub.memoize_property
-    def idx_to_id(self):
-        """
-        Example:
-            >>> import ndsampler
-            >>> self = ndsampler.CategoryTree.demo()
-            >>> self.idx_to_id[0]
-        """
-        return [self.node_to_id[node]
-                for node in self.idx_to_node]
-
-    @ub.memoize_method
-    def idx_to_ancestor_idxs(self, include_self=True):
-        """
-        Mapping from a class index to its ancestors
-
-        Args:
-            include_self (bool, default=True):
-                if True includes each node as its own ancestor.
-        """
-        lut = {
-            idx: set(ub.take(self.node_to_idx, nx.ancestors(self.graph, node)))
-            for idx, node in enumerate(self.idx_to_node)
-        }
-        if include_self:
-            for idx, idxs in lut.items():
-                idxs.update({idx})
-        return lut
-
-    @ub.memoize_method
-    def idx_to_descendants_idxs(self, include_self=False):
-        """
-        Mapping from a class index to its descendants (including itself)
-
-        Args:
-            include_self (bool, default=False):
-                if True includes each node as its own descendant.
-        """
-        lut = {
-            idx: set(ub.take(self.node_to_idx, nx.descendants(self.graph, node)))
-            for idx, node in enumerate(self.idx_to_node)
-        }
-        if include_self:
-            for idx, idxs in lut.items():
-                idxs.update({idx})
-        return lut
-
-    @ub.memoize_method
-    def idx_pairwise_distance(self):
-        """
-        Get a matrix encoding the distance from one class to another.
-
-        Distances
-            * from parents to children are positive (descendants),
-            * from children to parents are negative (ancestors),
-            * between unreachable nodes (wrt to forward and reverse graph) are
-            nan.
-        """
-        pdist = np.full((len(self), len(self)), fill_value=-np.nan,
-                        dtype=np.float32)
-        for node1, dists in nx.all_pairs_shortest_path_length(self.graph):
-            idx1 = self.node_to_idx[node1]
-            for node2, dist in dists.items():
-                idx2 = self.node_to_idx[node2]
-                pdist[idx1, idx2] = dist
-                pdist[idx2, idx1] = -dist
-        return pdist
-
-    def __len__(self):
-        return len(self.graph)
-
-    def __iter__(self):
-        return iter(self.idx_to_node)
-
-    def __getitem__(self, index):
-        return self.idx_to_node[index]
-
-    def __json__(self):
-        """
-        Example:
-            >>> import pickle
-            >>> self = CategoryTree.demo()
-            >>> print('self = {!r}'.format(self.__json__()))
-        """
-        return self.__getstate__()
-
-    def __getstate__(self):
-        """
-        Serializes information in this class
-
-        Example:
-            >>> from ndsampler.category_tree import *
-            >>> import pickle
-            >>> self = CategoryTree.demo()
-            >>> state = self.__getstate__()
-            >>> serialization = pickle.dumps(self)
-            >>> recon = pickle.loads(serialization)
-            >>> assert recon.__json__() == self.__json__()
-        """
-        state = self.__dict__.copy()
-        for key in list(state.keys()):
-            if key.startswith('_cache'):
-                state.pop(key)
-        state['graph'] = to_directed_nested_tuples(self.graph)
-        if True:
-            # Remove reundant items
-            state.pop('node_to_idx')
-            state.pop('node_to_id')
-            state.pop('idx_groups')
-        return state
-
-    def __setstate__(self, state):
-        graph = from_directed_nested_tuples(state['graph'])
-
-        if True:
-            # Reconstruct redundant items
-            if 'node_to_idx' not in state:
-                state['node_to_idx'] = {node: idx for idx, node in
-                                        enumerate(state['idx_to_node'])}
-            if 'node_to_id' not in state:
-                state['node_to_id'] = {node: id for id, node in
-                                       state['id_to_node'].items()}
-
-            if 'idx_groups' not in state:
-                node_groups = list(traverse_siblings(graph))
-                node_to_idx = state['node_to_idx']
-                state['idx_groups'] = [sorted([node_to_idx[n] for n in group])
-                                       for group in node_groups]
-
-        self.__dict__.update(state)
-        self.graph = graph
-
-    def __nice__(self):
-        return 'nNodes={}, maxDepth={}, maxBreadth={}, nodes={}'.format(
-            self.num_classes,
-            tree_depth(self.graph),
-            max(it.chain([0], map(len, self.idx_groups))),
-            self.idx_to_node,
-        )
-
-    def _demo_probs(self, num=5, rng=0, nonrandom=3, hackargmax=True):
-        """ dummy probabilities for testing """
-        rng = kwarray.ensure_rng(rng)
-        class_energy = torch.FloatTensor(rng.rand(num, len(self)))
-
-        # Setup the first few examples to prefer being classified
-        # as a fine grained class to a decreasing degree.
-        # The first example is set to have equal energy
-        # The i + 2-th example is set to have an extremely high energy.
-        start = 0
-        nonrandom = min(nonrandom, (num - start))
-        if nonrandom > 0:
-            path = sorted(ub.take(self.node_to_idx, nx.dag_longest_path(self.graph)))
-
-            class_energy[start] = 1 / len(class_energy[start])
-            if hackargmax:
-                # HACK: even though we want to test uniform distributions, it makes
-                # regression tests difficiult because torch and numpy return a
-                # different argmax when the array has more than one max value.
-                # add a VERY small epsilon to make max values distinct
-                class_energy[start] += torch.linspace(0, .00001, len(class_energy[start]))
-
-            if nonrandom > 1:
-                for i in range(nonrandom - 2):
-                    class_energy[start + i + 1][path] += 2 ** (i / 4)
-                class_energy[start + i + 2][path] += 2 ** 20
-
-        class_probs = self.hierarchical_softmax(class_energy, dim=1)
-        return class_probs
-
-    @classmethod
-    def demo(cls, key='coco', **kwargs):
-        """
-        Args:
-            key (str): specify which demo dataset to use.
-
-        CommandLine:
-            xdoctest -m ~/code/ndsampler/ndsampler/category_tree.py CategoryTree.demo
-
-        Example:
-            >>> from ndsampler.category_tree import *
-            >>> self = CategoryTree.demo()
-            >>> print('self = {}'.format(self))
-            self = <CategoryTree(nNodes=10, maxDepth=2, maxBreadth=4...)>
-        """
-        if key == 'coco':
-            from ndsampler import coco_dataset
-            dset = coco_dataset.CocoDataset.demo(**kwargs)
-            dset.add_category('background', id=0)
-            graph = dset.category_graph()
-        elif key == 'btree':
-            r = kwargs.pop('r', 3)
-            h = kwargs.pop('h', 3)
-            graph = nx.generators.balanced_tree(r=r, h=h, create_using=nx.DiGraph())
-            graph = nx.relabel_nodes(graph, {n: n + 1 for n in graph})
-            if kwargs.pop('add_zero', True):
-                graph.add_node(0)
-            assert not kwargs
-        else:
-            raise KeyError(key)
-        self = cls(graph)
-        return self
-
-    def to_coco(self):
-        """
-        Converts to a coco-style data structure
-        """
-        for cid, node in self.id_to_node.items():
-            # Skip if background already added
-            cat = {
-                'id': cid,
-                'name': node,
-            }
-            parents = list(self.graph.predecessors(node))
-            if len(parents) == 1:
-                cat['supercategory'] = parents[0]
-            else:
-                if len(parents) > 1:
-                    raise Exception('not a tree')
-            yield cat
-
-    def is_mutex(self):
-        """
-        Returns True if all categories are mutually exclusive (i.e. flat)
-
-        If true, then the classes may be represented as a simple list of class
-        names without any loss of information, otherwise the underlying
-        category graph is necessary to preserve all knowledge.
-
-        TODO:
-            - [ ] what happens when we have a dummy root?
-        """
-        return len(self.graph.edges) == 0
-
-    @property
-    def num_classes(self):
-        return self.graph.number_of_nodes()
-
-    @property
-    def class_names(self):
-        return self.idx_to_node
-
-    @property
-    def category_names(self):
-        return self.idx_to_node
-
-    @property
-    def cats(self):
-        """
-        Returns a mapping from category names to category attributes.
-
-        If this category tree was constructed from a coco-dataset, then this
-        will contain the coco category attributes.
-
-        Returns:
-            Dict[str, Dict[str, object]]
-
-        Example:
-            >>> from ndsampler.category_tree import *
-            >>> self = CategoryTree.demo()
-            >>> print('self.cats = {!r}'.format(self.cats))
-        """
-        return dict(self.graph.nodes)
-
-    def index(self, node):
-        """ Return the index that corresponds to the category name """
-        return self.node_to_idx[node]
-
-    def _build_index(self):
-        """ construct lookup tables """
-        # Most of the categories should have been given integer ids
-        max_id = max(it.chain([0], nx.get_node_attributes(self.graph, 'id').values()))
-        # Fill in id-values for any node that doesn't have one
-        node_to_id = {}
-        for node, attrs in sorted(self.graph.nodes.items()):
-            node_to_id[node] = attrs.get('id', max_id + 1)
-            max_id = max(max_id, node_to_id[node])
-        id_to_node = ub.invert_dict(node_to_id)
-
-        # Compress ids into a flat index space (sorted by node ids)
-        idx_to_node = ub.argsort(node_to_id)
-        node_to_idx = {node: idx for idx, node in enumerate(idx_to_node)}
-
-        # Find the sets of nodes that need to be softmax-ed together
-        node_groups = list(traverse_siblings(self.graph))
-        idx_groups = [sorted([node_to_idx[n] for n in group])
-                      for group in node_groups]
-
-        # Set instance attributes
-        self.id_to_node = id_to_node
-        self.node_to_id = node_to_id
-        self.idx_to_node = idx_to_node
-        self.node_to_idx = node_to_idx
-        self.idx_groups = idx_groups
 
     def conditional_log_softmax(self, class_energy, dim):
         """
@@ -591,13 +129,15 @@ class CategoryTree(ub.NiceRepr):
 
     def source_log_softmax(self, class_energy, dim):
         """
-        Top-down hierarchical softmax
+        Top-down hierarchical softmax. This is the default
+        hierarchical_log_softmax function.
 
         Alternative to `sink_log_softmax`
 
         SeeAlso:
             * sink_log_softmax
-            * hierarchical_log_softmax
+            * hierarchical_log_softmax (alias for this function)
+            * source_log_softmax (this function)
 
         Converts raw class energy to absolute log probabilites based on the
         category hierarchy. This is done by first converting to conditional
@@ -635,17 +175,24 @@ class CategoryTree(ub.NiceRepr):
 
     def sink_log_softmax(self, class_energy, dim):
         """
-        Bottom-up hierarchical softmax
+        Bottom-up hierarchical softmax.
 
         Alternative to `source_log_softmax`
 
         SeeAlso:
+            * sink_log_softmax (this function)
+            * hierarchical_log_softmax (alias for source log softmax)
             * source_log_softmax
-            * hierarchical_log_softmax
 
         Does a regular softmax over all the mutually exclusive leaf nodes, then
         sums their probabilities to get the score for the parent nodes.
 
+        Notes:
+            In this method of computation ONLY the energy in the leaf nodes
+            matters. The energy in all other intermediate notes is ignored.
+            For this reason the "source" method of computation is often
+            prefered.
+
         Args:
             class_energy (Tensor): raw output from network. The values in
                 `class_energy[..., idx]` should correspond to the network
@@ -799,27 +346,6 @@ class CategoryTree(ub.NiceRepr):
         loss = F.nll_loss(class_logprobs, targets)
         return loss
 
-    def show(self):
-        """
-        Ignore:
-            >>> import kwplot
-            >>> kwplot.autompl()
-            >>> from ndsampler import category_tree
-            >>> self = category_tree.CategoryTree.demo()
-            >>> self.show()
-
-            python -c "import kwplot, ndsampler, graphid; kwplot.autompl(); graphid.util.show_nx(ndsampler.category_tree.CategoryTree.demo().graph); kwplot.show_if_requested()" --show
-        """
-        try:
-            pos = nx.drawing.nx_agraph.graphviz_layout(self.graph, prog='dot')
-        except ImportError:
-            import warnings
-            warnings.warn('pygraphviz is not available')
-            pos = None
-        nx.draw_networkx(self.graph, pos=pos)
-        # import graphid
-        # graphid.util.show_nx(self.graph)
-
     def _prob_decision(self, class_probs, dim, thresh=0.1):
         """
         Chooses the finest-grained category based on raw prob threshold
@@ -906,6 +432,36 @@ class CategoryTree(ub.NiceRepr):
         pred_idxs, pred_conf = _descend(0, nodes, jdxs)
         return pred_idxs, pred_conf
 
+    def _demo_probs(self, num=5, rng=0, nonrandom=3, hackargmax=True):
+        """ dummy probabilities for testing """
+        rng = kwarray.ensure_rng(rng)
+        class_energy = torch.FloatTensor(rng.rand(num, len(self)))
+
+        # Setup the first few examples to prefer being classified
+        # as a fine grained class to a decreasing degree.
+        # The first example is set to have equal energy
+        # The i + 2-th example is set to have an extremely high energy.
+        start = 0
+        nonrandom = min(nonrandom, (num - start))
+        if nonrandom > 0:
+            path = sorted(ub.take(self.node_to_idx, nx.dag_longest_path(self.graph)))
+
+            class_energy[start] = 1 / len(class_energy[start])
+            if hackargmax:
+                # HACK: even though we want to test uniform distributions, it makes
+                # regression tests difficiult because torch and numpy return a
+                # different argmax when the array has more than one max value.
+                # add a VERY small epsilon to make max values distinct
+                class_energy[start] += torch.linspace(0, .00001, len(class_energy[start]))
+
+            if nonrandom > 1:
+                for i in range(nonrandom - 2):
+                    class_energy[start + i + 1][path] += 2 ** (i / 4)
+                class_energy[start + i + 2][path] += 2 ** 20
+
+        class_probs = self.hierarchical_softmax(class_energy, dim=1)
+        return class_probs
+
     def decision(self, class_probs, dim, thresh=0.5, criterion='gini',
                  ignore_class_idxs=None, always_refine_idxs=None):
         """
@@ -1014,6 +570,38 @@ class CategoryTree(ub.NiceRepr):
             >>> class_probs = self._demo_probs(num=10, nonrandom=8)
             >>> pred_idxs1, pref_conf1 = self.decision(class_probs, dim=1)
             >>> print('pred_idxs1 = {!r}'.format(pred_idxs1))
+
+        Ignore:
+            >>> import kwcoco
+            >>> from kwcoco.category_tree import _print_forest
+            >>> dset = kwcoco.CocoDataset.demo('shapes8')
+            >>> self = dset.object_categories()
+            >>> _print_forest(self.graph)
+            ├── background
+            ├── raster
+            │   ├── eff
+            │   └── superstar
+            └── vector
+                └── star
+
+            >>> print(list(self))
+            ['background', 'star', 'superstar', 'eff', 'raster', 'vector']
+
+            >>> class_probs = np.array([[0.05, 0.10, 0.90, 0.10, .90, .05]])
+            >>> self.decision(class_probs, dim=1)
+            >>> print('node = {!r}, prob = {!r}'.format(self.idx_to_node[idx[0]], prob[0]))
+
+            >>> class_probs = np.array([[0.00, 0.00, 0.98, 0.02, 0.98, 0.02]])
+            >>> idx, prob = self.decision(class_probs, dim=1, thresh=0.1, criterion='entropy')
+            >>> print('node = {!r}, prob = {!r}'.format(self.idx_to_node[idx[0]], prob[0]))
+
+            >>> class_probs = np.array([[0.00, 0.98, 0.01, 0.01, 0.02, 0.98]])
+            >>> idx, prob = self.decision(class_probs, dim=1, thresh=0.01, criterion='entropy')
+            >>> print('node = {!r}, prob = {!r}'.format(self.idx_to_node[idx[0]], prob[0]))
+
+            >>> class_probs = np.array([[0.00, 1.00, 0.00, 0.00, 0.00, 1.00]])
+            >>> idx, prob = self.decision(class_probs, dim=1, thresh=0.1, criterion='entropy')
+            >>> print('node = {!r}, prob = {!r}'.format(self.idx_to_node[idx[0]], prob[0]))
         """
         if criterion == 'prob':
             return self._prob_decision(class_probs, dim, thresh=thresh)
@@ -1132,20 +720,31 @@ class CategoryTree(ub.NiceRepr):
                         if idx in always_refine_idxs:
                             refine_flags[:] = 1
 
+                    if len(child_idxs) == 1:
+                        # hack: always refine when there is one child, in this
+                        # case the simplicity measure will always be zero,
+                        # which is likely a problem with this criterion.
+                        refine_flags[:] = 1
+
                     refine_flags = kwarray.ArrayAPI.numpy(refine_flags).astype(np.bool)
 
                     if DEBUG:
                         print('-----------')
                         print('idx = {!r}'.format(idx))
+                        print('node = {!r}'.format(self.idx_to_node[idx]))
                         print('ommer_idxs = {!r}'.format(ommer_idxs))
+                        print('ommer_nodes = {!r}'.format(
+                            list(ub.take(self.idx_to_node, ommer_idxs))))
                         print('depth = {!r}'.format(depth))
                         import pandas as pd
                         print('expanded_probs =\n{}'.format(
-                            ub.repr2(expanded_probs, precision=2, supress_small=True)))
+                            ub.repr2(expanded_probs, precision=2,
+                                     with_dtype=0, supress_small=True)))
                         df = pd.DataFrame({
                             'h': h_expanded,
                             'h_worst': h_expanded_worst,
-                            'ratio': complexity_ratio,
+                            'c_ratio': complexity_ratio,
+                            's_ratio': simplicity_ratio,
                             'flags': refine_flags.astype(np.uint8)
                         })
                         print(df)
@@ -1189,6 +788,11 @@ class CategoryTree(ub.NiceRepr):
         return self.hierarchical_log_softmax(*args, **kw)
 
 
+class CategoryTree(KWCOCO_CategoryTree, Mixin_CategoryTree_Torch):
+    # Mixin the kwcoco category tree with torch functionality
+    KWCOCO_CategoryTree.__doc__
+
+
 def source_nodes(graph):
     """ generates source nodes --- nodes without incoming edges """
     return (n for n in graph.nodes() if graph.in_degree(n) == 0)
@@ -1199,105 +803,6 @@ def sink_nodes(graph):
     return (n for n in graph.nodes() if graph.out_degree(n) == 0)
 
 
-def traverse_siblings(graph, sources=None):
-    """ generates groups of nodes that have the same parent """
-    if sources is None:
-        sources = list(source_nodes(graph))
-    yield sources
-    for node in sources:
-        children = list(graph.successors(node))
-        if children:
-            for _ in traverse_siblings(graph, children):
-                yield _
-
-
-def tree_depth(graph, root=None):
-    """
-    Maximum depth of the forest / tree
-
-    Example:
-        >>> from ndsampler.category_tree import *
-        >>> graph = nx.balanced_tree(r=2, h=3, create_using=nx.DiGraph)
-        >>> tree_depth(graph)
-        4
-        >>> tree_depth(nx.balanced_tree(r=2, h=0, create_using=nx.DiGraph))
-        1
-    """
-    if len(graph) == 0:
-        return 0
-    if root is not None:
-        assert root in graph.nodes
-    assert nx.is_forest(graph)
-    def _inner(root):
-        if root is None:
-            return max(it.chain([0], (_inner(n) for n in source_nodes(graph))))
-        else:
-            return max(it.chain([0], (_inner(n) for n in graph.successors(root)))) + 1
-    depth = _inner(root)
-    return depth
-
-
-def to_directed_nested_tuples(graph, with_data=True):
-    """
-    Encodes each node and its children in a tuple as:
-        (node, children)
-    """
-    def _represent_node(node):
-        if with_data:
-            node_data = graph.nodes[node]
-            return (node, node_data, _traverse_encode(node))
-        else:
-            return (node, _traverse_encode(node))
-
-    def _traverse_encode(parent):
-        children = sorted(graph.successors(parent))
-        # graph.get_edge_data(node, child)
-        return [_represent_node(node) for node in children]
-
-    sources = sorted(source_nodes(graph))
-    encoding = [_represent_node(node) for node in sources]
-    return encoding
-
-
-def from_directed_nested_tuples(encoding):
-    """
-    Example:
-        >>> from ndsampler.category_tree import *
-        >>> graph = nx.generators.gnr_graph(20, 0.3, seed=790).reverse()
-        >>> graph.nodes[0]['color'] = 'black'
-        >>> encoding = to_directed_nested_tuples(graph)
-        >>> recon = from_directed_nested_tuples(encoding)
-        >>> recon_encoding = to_directed_nested_tuples(recon)
-        >>> assert recon_encoding == encoding
-    """
-    node_data_view = {}
-    def _traverse_recon(tree):
-        nodes = []
-        edges = []
-        for tup in tree:
-            if len(tup) == 2:
-                node, subtree = tup
-            elif len(tup) == 3:
-                node, node_data, subtree = tup
-                node_data_view[node] = node_data
-            else:
-                raise AssertionError('invalid tup')
-            children = [t[0] for t in subtree]
-            nodes.append(node)
-            edges.extend((node, child) for child in children)
-            subnodes, subedges = _traverse_recon(subtree)
-            nodes.extend(subnodes)
-            edges.extend(subedges)
-        return nodes, edges
-    nodes, edges = _traverse_recon(encoding)
-    graph = nx.DiGraph()
-    graph.add_nodes_from(nodes)
-    graph.add_edges_from(edges)
-    for k, v in node_data_view.items():
-        graph.nodes[k].update(v)
-    return graph
-
-
 def gini(probs, axis=1, impl=np):
     """
     Approximates Shannon Entropy, but faster to compute
@@ -1332,24 +837,6 @@ def entropy(probs, axis=1, impl=np):
         return h
 
 
-class _calldict(dict):
-    """
-    helper object to maintain backwards compatibility between new and old
-    id_to_idx methods.
-
-    Example:
-        >>> self = _calldict({1: 2})
-        >>> #assert self()[1] == 2
-        >>> assert self[1] == 2
-    """
-
-    def __call__(self):
-        import warnings
-        warnings.warn('Calling id_to_idx as a method has been depricated. '
-                      'Use this dict as a property')
-        return self
-
-
 if __name__ == '__main__':
     """
     CommandLine:
